from typing import Dict, Any

from sktmls import MLSClient, MLSENV, MLSRuntimeENV, MLSResponse

MLS_DYNAMODB_API_URL = "/api/v1/dynamodb"


class DynamoDBClient(MLSClient):
    """
    DynamoDB 관련 기능들을 제공하는 클라이언트입니다.
    위 클라이언트는 STG 환경만 지원됩니다.
    """

    def __init__(
        self,
        env: MLSENV = None,
        runtime_env: MLSRuntimeENV = None,
        username: str = None,
        password: str = None,
    ):
        """
        ## Args

        - env: (`sktmls.MLSENV`) 접근할 MLS 환경. 본 클라이언트는 STG 환경(`sktmls.MLSENV.STG`)만 지원
        - runtime_env: (`sktmls.MLSRuntimeENV`) 클라이언트가 실행되는 환경 (`sktmls.MLSRuntimeENV.YE`|`sktmls.MLSRuntimeENV.EDD`|`sktmls.MLSRuntimeENV.LOCAL`) (기본값: `sktmls.MLSRuntimeENV.LOCAL`)
        - username: (str) MLS 계정명 (기본값: $MLS_USERNAME)
        - password: (str) MLS 계정 비밀번호 (기본값: $MLS_PASSWORD)

        ## Returns
        `sktmls.dynamodb.DynamoDBClient`

        ## Example

        ```python
        dynamodb_client = DynamoDBClient(env=MLSENV.STG, runtime_env=MLSRuntimeENV.YE, username="mls_account", password="mls_password")
        ```
        """
        super().__init__(env=env, runtime_env=runtime_env, username=username, password=password)
        assert self._MLSClient__env == MLSENV.STG, "STG 환경만 지원합니다"

    def get_item(
        self,
        table_name: str,
        key: Dict[str, Any],
    ) -> Dict[str, Any]:
        """
        주어진 Dynamodb 테이블의 특정 아이템을 조회합니다

        ## Args

        - table_name: (str) 테이블 명
        - key: (dict) 조회할 아이템의 key 값

        ## Returns

        (dict) 해당 아이템 값

        ## Example

        ```python
        item = dynamodb_client.get_item(table_name="sample_table", key={"item_id": "sample"})
        ```
        """
        return self._request(
            method="POST", url=MLS_DYNAMODB_API_URL, data={"type": "get", "table_name": table_name, "body": key}
        ).results

    def put_item(self, table_name: str, item: Dict[str, Any]) -> Dict[str, Any]:
        """
        주어진 Dynamodb 테이블에 특정 아이템을 추가합니다

        ## Args

        - table_name: (str) 테이블 명
        - item: (dict) 추가할 아이템 값

        ## Returns

        (dict) 추가한 아이템 값

        ## Example
        ```python
        dynamodb_client.put_item(
            table_name="sample_table",
            item={"item_id": "sample", "sample_attribute": "sample_value"}
        )
        ```
        """
        return self._request(
            method="POST",
            url=MLS_DYNAMODB_API_URL,
            data={"type": "put", "table_name": table_name, "body": item},
        ).results

    def update_item(self, table_name: str, item: Dict[str, Any]) -> Dict[str, Any]:
        """
        주어진 Dynamodb 테이블의 특정 아이템을 수정합니다

        ## Args

        - table_name: (str) 테이블 명
        - item: (dict) 수정할 아이템 값

        ## Returns

        (dict) 수정한 아이템 값

        ## Example

        ```python
        dynamodb_client.update_item(
            table_name="sample_table",
            item={"item_id": "sample", "sample_attribute": "sample_value_updated", "add_new_field": "added_field_value"}
        )
        ```
        """
        return self._request(
            method="POST",
            url=MLS_DYNAMODB_API_URL,
            data={"type": "update", "table_name": table_name, "body": item},
        ).results

    def delete_item(self, table_name: str, key: Dict[str, Any]) -> MLSResponse:
        """
        주어진 Dynamodb 테이블의 특정 아이템을 삭제합니다

        ## Args

        - table_name: (str) 테이블 명
        - key: (dict) 삭제할 아이템의 key 값

        ## Returns

        `sktmls.MLSResponse`

        ## Example

        ```python
        dynamodb_client.delete_item(
            table_name="sample_table",
            item={"item_id": "sample"}
        )
        ```
        """
        return self._request(
            method="POST",
            url=MLS_DYNAMODB_API_URL,
            data={"type": "delete", "table_name": table_name, "body": key},
        )



from typing import Any, Dict, List, Union

import numpy as np
from pandas import DataFrame
from pytz import timezone

from sktmls import MLSRuntimeENV
from sktmls.apis import MLSProfileAPIClient
from sktmls.models import MLSGenericModel, MLSModelError, MLSTrainable
from sktmls.utils import LogicProcessor

TZ = timezone("Asia/Seoul")

logic_processor = LogicProcessor()


class GenericLogicModel(MLSGenericModel, MLSTrainable):
    """
    MLS 모델 레지스트리에 등록되는 단일 모델 기반의 클래스입니다.

    전처리 로직과 후처리 로직을 json 형태로 전달하여 프로세스합니다.
    """

    def __init__(
        self,
        model_name: str,
        model_version: str,
        features: List[str],
        model=None,
        preprocess_logic: Dict[str, List[Any]] = None,
        postprocess_logic: Dict[str, List[Any]] = None,
        predict_fn: str = "predict",
        data: Dict[str, Any] = {},
    ):
        """
        ## Args

        - model_name: (str) 모델 이름
        - model_version: (str) 모델 버전
        - features: (list(str)) 피쳐 이름 리스트
        - model: (optional) ML 라이브러리로 학습한 모델 객체 (기본값: None)
        - preprocess_logic: (optional) (dict) 전달된 피쳐 `x`로부터 ML모델의 predict 함수에 input으로 들어갈 `preprocessed_x`를 만드는 전처리 로직 (기본값: 아래 참조)
        - postprocess_logic: (optional) (dict) ML모델의 predict 함수 결과로 얻어진 `y`로부터 리턴 body(items 리스트)를 만드는 후처리 로직 (기본값: 아래 참조)
        - predict_fn: (optional) (str) ML모델의 추론 함수 이름 (`predict`|`predict_proba`|`none`) (기본값: `predict`)
            - `none` 전달 시 ML모델 추론을 사용하지 않습니다 (룰 모델).
        - data: (optional) (dict) `preprocess_logic`과 `postprocess_logic`에서 피쳐와 함께 "var" 참조할 추가 데이터 (기본값: {})
            - 피쳐 이름과 같은 키 존재 시 피쳐 값을 덮어쓰게 됩니다. 주의하세요!

        ## Example

        자세한 예제는 아래를 참조하세요.

        - https://github.com/sktaiflow/mls-samples/tree/main/GenericLogicModel


        ```python

        # 피쳐 이름
        # 유저 프로파일에 존재하는 디멘전의 이름 리스트
        features = ["feature1", "feature2", "feature3", "embedding_vector", "context_feature1", "context_feature2"]

        # 전처리 로직
        # 전달된 피쳐로부터 preprocessed_x를 만드는 로직을 정의합니다.
        # 피쳐 값은 {"var": ["피쳐이름", 기본값]} 으로 참조합니다.
        # 이해를 돕기 위한 예시이며 실제 문제와는 다를 수 있습니다. 그대로 사용하지 마시고 참고만 해 주세요.
        # 전달하지 않은 경우 x 리스트를 전처리 없이 그대로 사용합니다.
        preprocess_logic = {
            # float: 리스트의 모든 element를 float으로 캐스팅합니다.
            "float": [
                # merge: element들을 하나의 리스트에 합쳐 반환합니다.
                {"merge": [
                    {"var": ["feature1", 0]},
                    {"if": [
                        {"==": [{"var": ["feature2", "N/A"]}, "S"]},
                        1,
                        {"==": [{"var": ["feature2", "N/A"]}, "V"]},
                        2,
                        0
                    ]},
                    {"var": ["feature3", 0]},
                    # replace: 리스트 내 None을 0.0으로 교체
                    {"replace": [
                        None,
                        0.0,
                        # pick: 리스트의 특정 인덱스만 뽑아서 리턴
                        {"pick": [
                            {"var": ["embedding_vector", [0.0] * 64]},
                            [1, 2]
                        ]}
                    ]},
                    # weekday: 요일 리턴 (월요일: 1 ~ 일요일: 7)
                    {"weekday": []},
                    # day: 오늘 날짜
                    # ndays: 이번 달 마지막 날짜
                    # /: 나누기
                    {"/": [{"day": []}, {"ndays": []}]},
                    # get: 리스트 또는 딕셔너리에서 해당 인덱스 꺼내기
                    {"get": [
                        {"replace": [
                            "#",
                            "0.0",
                            {
                                # pf: 프로파일 조회
                                # - 타입: (`user`|`item`)
                                # - 프로파일ID
                                # - 조회할 user_id 또는 item_id
                                # - 조회할 키 디멘전
                                "pf": [
                                    "item",
                                    "device",
                                    # additional_keys: Recommendation API(v3 이상)으로부터 전달된 추가 키 리스트
                                    # additional_keys.0: 0번째 element
                                    {"var": ["additional_keys.0", "ABCD"]},
                                    ["sale_cnt"]
                                ]
                            }
                        ]},
                        0
                    ]}
                ]
            }]
        }

        # 후처리 로직
        # 계산된 y로부터 최종 리턴 body(items 리스트)를 만드는 로직을 정의합니다.
        # 피쳐 값은 {"var": "피쳐이름"} 으로 참조합니다.
        # y 값은 {"var": "y"} 으로 참조합니다.
        # y가 리스트인 경우 {"var": "y.0"}, {"var": "y.1"} 등으로 참조합니다.
        # 이해를 돕기 위한 예시이며 실제 문제와는 다를 수 있습니다. 그대로 사용하지 마시고 참고만 해 주세요.
        # 전달하지 않은 경우 아래 기본값이 사용됩니다.
        # [{"id": "ID", "name": "NAME", "type": "TYPE", "props": {"score": {"var": "y"}}}]
        postprocess_logic = {
            "if": [
                # if
                # 25 <= age < 65 이며
                # five_g_yn == "N" 인 경우
                {
                    "and": [
                        {">=": [{"var": ["age", 0]}, 25]},
                        {"<": [{"var": ["age", 0]}, 65]},
                        {"==": [{"var": ["five_g_yn", "N"]}, "N"]},
                    ]
                },
                # then
                {
                    "list": [
                        {
                            "dict": [
                                "id", "PRODUCT001",
                                "name", "상품001",
                                "type", "타입",
                                "props", {
                                    "dict": [
                                        "context_id", {
                                            "if": [
                                                # if context_feature1의 값이 Y이면
                                                {"==": [{"var": ["context_feature1", "N"]}, "Y"]},
                                                # "context_feature1"을 context_id로 사용
                                                "context_feature1",
                                                # else if context_feature2의 값이 Y이면
                                                {"==": [{"var": ["context_feature2", "N"]}, "Y"]},
                                                # "context_feature2"을 context_id로 사용
                                                "context_feature2",
                                                # else "default_context"를 context_id로 사용
                                                "default_context"
                                            ]
                                        }
                                    ]
                                }
                            ]
                        }
                    ]
                },
                # else if
                # 19 <= age < 25 이며
                # five_g_yn == "N" 인 경우
                {
                    "and": [
                        {">=": [{"var": ["age", 0]}, 19]},
                        {"<": [{"var": ["age", 0]}, 25]},
                        {"==": [{"var": ["five_g_yn", "N"]}, "N"]},
                    ]
                },
                # then
                {
                    "list": [
                        {
                            "dict": [
                                "id", "PRODUCT002",
                                "name", "상품002",
                                "type", "타입",
                                "props", {
                                    "dict": [
                                        "context_id", {
                                            "if": [
                                                # if context_feature1의 값이 Y이면
                                                {"==": [{"var": ["context_feature1", "N"]}, "Y"]},
                                                # "context_feature1"을 context_id로 사용
                                                "context_feature1",
                                                # else if context_feature2의 값이 Y이면
                                                {"==": [{"var": ["context_feature2", "N"]}, "Y"]},
                                                # "context_feature2"을 context_id로 사용
                                                "context_feature2",
                                                # else "default_context"를 context_id로 사용
                                                "default_context"
                                            ]
                                        }
                                    ]
                                }
                            ]
                        }
                    ]
                },
                # else
                # None을 리턴
                None
            ]
        }

        my_model_v1 = GenericLogicModel(
            model=model,
            model_name="my_model",
            model_version="v1",
            features=features,
            preprocess_logic=preprocess_logic,
            postprocess_logic=postprocess_logic,
            predict_fn="predict"
        )

        # 정확한 로깅을 위한 사용 라이브러리 명시 (권장) (`sklearn`|`lightgbm`|`xgboost`|`catboost`|`rule`|`etc`)
        my_model_v1.model_lib = "lightgbm"
        my_model_v1.model_lib_version = "2.3.1"
        ```
        """
        assert isinstance(features, list), "`features`은 list 타입이어야 합니다."

        if preprocess_logic is not None:
            assert isinstance(preprocess_logic, dict), "`preprocess_logic`은 dict 타입이어야 합니다."
            for key in preprocess_logic.keys():
                assert (
                    key in ["var", "missing", "missing_some", "pf", "with", "for"] or key in logic_processor.operations
                )
        else:
            preprocess_logic = {"merge": [{"var": f} for f in features]}

        if postprocess_logic is not None:
            assert isinstance(postprocess_logic, dict), "`postprocess_logic`은 dict 타입이어야 합니다."
            for key in postprocess_logic.keys():
                assert (
                    key in ["var", "missing", "missing_some", "pf", "with", "for"] or key in logic_processor.operations
                )
        else:
            postprocess_logic = {
                "list": [
                    {"dict": ["id", "id", "name", "name", "type", "type", "props", {"dict": ["score", {"var": "y"}]}]}
                ]
            }

        assert isinstance(predict_fn, str), "`predict_fn`은 str 타입이어야 합니다."
        assert predict_fn in [
            "predict",
            "predict_proba",
            "none",
        ], "`predict_fn`은 predict, predict_proba, none 중 하나의 값이어야 합니다."

        assert isinstance(data, dict), "`data`는 dict 타입이어야 합니다."

        super().__init__([model], model_name, model_version, features)

        self.preprocess_logic = preprocess_logic
        self.postprocess_logic = postprocess_logic
        self.predict_fn = predict_fn
        self.data = data

    def predict(self, x: List[Any], **kwargs) -> Dict[str, Any]:
        pf_client = kwargs.get("pf_client") or MLSProfileAPIClient(runtime_env=MLSRuntimeENV.MMS)

        preprocessed_x = self._preprocess(x, kwargs.get("keys", []), pf_client)
        y = self._ml_predict(preprocessed_x)
        items = self._postprocess(x, kwargs.get("keys", []), y, pf_client) or []

        return {"items": items}

    def _preprocess(self, x: List[Any], additional_keys: List[Any], pf_client: MLSProfileAPIClient) -> List[Any]:
        if len(self.features) != len(x):
            raise MLSModelError("GenericLogicModel: `x`의 길이가 `features`의 길이와 다릅니다.")

        data = {name: x[i] for i, name in enumerate(self.features) if x[i] not in [None, []]}
        data["additional_keys"] = additional_keys
        data.update(self.data)

        try:
            return logic_processor.apply(self.preprocess_logic, data=data, pf_client=pf_client)
        except Exception as e:
            raise MLSModelError(f"GenericLogicModel: 전처리에 실패했습니다. {e}")

    def _ml_predict(self, preprocessed_x: List[Any]) -> Union[float, List[float], str, None]:
        try:
            if self.predict_fn == "none":
                return None

            if not isinstance(preprocessed_x[0], list):
                preprocessed_x = [preprocessed_x]

            if self.model_lib == "autogluon":
                input_data = DataFrame(
                    preprocessed_x, columns=[f for f in self.features if f not in self.non_training_features]
                )
            else:
                input_data = np.array(preprocessed_x)

            if self.predict_fn == "predict":
                y = self.models[0].predict(input_data)
            else:
                y = self.models[0].predict_proba(input_data)

            if len(y) == 1:
                y = y[0]

            try:
                return y.tolist()
            except AttributeError:
                return y

        except Exception as e:
            raise MLSModelError(f"GenericLogicModel: ML Prediction에 실패했습니다. {e}")

    def _postprocess(
        self,
        x: List[Any],
        additional_keys: List[Any],
        y: Union[float, List[float], None],
        pf_client: MLSProfileAPIClient,
    ) -> List[Dict[str, Any]]:

        data = {name: x[i] for i, name in enumerate(self.features) if x[i] not in [None, []]}
        data["additional_keys"] = additional_keys
        data["y"] = y
        data.update(self.data)

        try:
            return logic_processor.apply(self.postprocess_logic, data=data, pf_client=pf_client)
        except Exception as e:
            raise MLSModelError(f"GenericLogicModel: 후처리에 실패했습니다. {e}")



from abc import ABC, abstractmethod
from typing import List, Iterator, Iterable, Tuple, Type, Optional
import mxnet as mx
import numpy as np
from mxnet.ndarray import NDArray

from ..datatypes.common import Candidate
from ..datatypes.tuning_job_state import TuningJobState


class NextCandidatesAlgorithm:
    def next_candidates(self) -> List[Candidate]:
        # Not using ABC otherwise it will be difficult to create a subclass that also is a
        # NamedTuple(as both define their own metaclass)
        raise NotImplemented("Abstract method")


class CandidateGenerator(ABC):
    """
    Class to generate candidates from which to start the local minimization, typically random candidate
    or some form of more uniformly spaced variation, such as latin hypercube or sobol sequence
    """
    @abstractmethod
    def generate_candidates(self) -> Iterator[Candidate]:
        pass

    def generate_candidates_en_bulk(self, num_cands: int) -> List[Candidate]:
        raise NotImplementedError()


class SurrogateModel(ABC):
    def __init__(self, state: TuningJobState, active_metric: str,
                 random_seed: int):
        self.state = state
        self.random_seed = random_seed
        self.active_metric = active_metric

    @abstractmethod
    def predict_nd(self, x_nd: NDArray) -> List[Tuple[NDArray, NDArray]]:
        """
        Given a (n, d) matrix x_nd of test input points, return predictive means
        and predictive stddevs, as (n,) vectors.

        Note: Different to state passed at construction, the input points in
        x_nd must be encoded already. See also 'predict_candidates'.

        If the model supports fantasizing (see FantasizingSurrogateModel), and
        the state passed at construction contains pending evaluations with
        fantasized target values, then pred_mean will be a matrix of shape
        (n, num_fantasy_samples), one column per fantasy sample, while pred_std
        remains a vector (the same for each sample).

        When using GP with marginal likelihood estimation, the returned list
        will be of length 1. When using GP with MCMC, the returned list will have
        one entry per MCMC sample.

        :param x_nd: Test input points
        :return: A list of (pred_mean, pred_std)
        """
        pass

    @abstractmethod
    def predict(self, X: np.ndarray) -> List[Tuple[np.ndarray, np.ndarray]]:
        """
        Wrapper around predict_nd, where inputs and returns are np.ndarray.

        Note: Different to state passed at construction, the input points in
        X must be encoded already. See also 'predict_candidates'.
        """
        pass

    def predict_candidates(self, candidates: Iterable[Candidate]) -> \
            List[Tuple[np.ndarray, np.ndarray]]:
        """
        Convenience function to get a list of means and standard deviations
        of candidates.
        """
        return self.predict(self.state.hp_ranges.to_ndarray_matrix(candidates))

    @abstractmethod
    def current_best(self) -> np.ndarray:
        """
        Returns the so-called incumbent, to be used in acquisition functions
        such as expected improvement. This is the minimum of predictive means
        at all current candidate locations (both state.candidate_evaluations
        and state.pending_evaluations).
        Normally, a scalar is returned, but if the model supports fantasizing
        and the state contains pending evaluations, there is one incumbent
        per fantasy sample, so a vector is returned.

        NOTE: When using MCMC, we should really maintain one incumbent per MCMC
        sample (for the same reason as we do that for fantasies). This is
        currently not done.

        :return: Incumbent
        """
        pass

    @abstractmethod
    def context_for_nd(self) -> mx.Context:
        """
        :return: Context for mx.nd input/output arguments
        """
        pass

    @abstractmethod
    def dtype_for_nd(self):
        """
        :return: Datatype for mx.nd input/output arguments
        """
        pass

    def convert_np_to_nd(self, x: np.ndarray) -> NDArray:
        return mx.nd.array(
            x, ctx=self.context_for_nd(), dtype=self.dtype_for_nd())


class ScoringFunction(ABC):
    """
    Class to score candidates, typically combine an acquisition function with
    potentially Thompson sampling

    NOTE: it will be minimized, i.e. lower is better
    """
    @abstractmethod
    def score(self, candidates: Iterable[Candidate],
              model: Optional[SurrogateModel] = None) -> List[float]:
        """
        Requires multiple candidates, is this can be much quicker: we can use matrix operations

        lower is better
        """
        pass


class AcquisitionFunction(ScoringFunction):
    @abstractmethod
    def __init__(self, model: SurrogateModel):
        self.model = model

    @abstractmethod
    def compute_acq(self, x: np.ndarray,
                    model: Optional[SurrogateModel] = None) -> np.ndarray:
        pass

    @abstractmethod
    def compute_acq_with_gradients(
            self, x: np.ndarray,
            model: Optional[SurrogateModel] = None) -> \
            Tuple[np.ndarray, np.ndarray]:
        pass

    def score(self, candidates: Iterable[Candidate],
              model: Optional[SurrogateModel] = None) -> List[float]:
        if model is None:
            model = self.model
        x = model.state.hp_ranges.to_ndarray_matrix(candidates)
        return list(self.compute_acq(x, model=model))


class LocalOptimizer(ABC):
    """
    Class that tries to find a local candidate with a better score, typically using a local
    optimization method such as lbfgs. It would normally encapsulate an acquisition function and model

    """
    def __init__(self, state: TuningJobState, model: SurrogateModel,
                 acquisition_function_class: Type[AcquisitionFunction]):
        self.state = state
        self.model = model
        self.acquisition_function_class = acquisition_function_class

    @abstractmethod
    def optimize(self, candidate: Candidate,
                 model: Optional[SurrogateModel] = None) -> Candidate:
        """
        Run local optimization, starting from candidate.
        If model is given, it overrides self.model.

        :param candidate: Starting point
        :param model: See above
        :return: Candidate found by local optimization
        """
        pass


class PendingCandidateStateTransformer(ABC):
    """
    This concept is needed if HPO can deal with pending candidates, which
    remain unlabeled during further decision-making (e.g., batch decisions
    or asynchronous HPO). In this case, information for the pending candidate
    has to be added to the state.

    """
    @abstractmethod
    def append_candidate(self, candidate: Candidate):
        """
        Determines PendingEvaluation information for candidate and append to
        state.pending_evaluations.

        :param candidate: Novel pending candidate
        """
        pass

    @property
    @abstractmethod
    def state(self) -> TuningJobState:
        """
        :return: Current TuningJobState
        """
        pass

    @abstractmethod
    def model(self, **kwargs) -> SurrogateModel:
        """
        In general, the model is computed here on demand, based on the current
        state, unless the state has not changed since the last 'model' call.

        :return: Surrogate model for current state
        """
        pass




from ...constants import AG_ARGS
from ...models.lgb.hyperparameters.parameters import get_param_baseline_custom


# Returns list of models created by a custom name preset.
def get_preset_custom(name, problem_type, num_classes):
    if not isinstance(name, str):
        raise ValueError(f'Expected string value for custom model, but was given {name}')
    # Custom model
    if name == 'GBM':
        model = get_param_baseline_custom(problem_type, num_classes=num_classes)
        model[AG_ARGS] = dict(model_type='GBM', name_suffix='Custom', disable_in_hpo=True)
        return [model]
    else:
        raise ValueError(f'Unknown custom model preset: {name}')



import pytest

from sktmls.models import MLSModelError
from sktmls.models.contrib import MbrBoostparkBaselineModel


class TestMbrBoostparkBaselineModel:
    def test_000_mbr_baseline_1(self):
        mbm = MbrBoostparkBaselineModel(
            model_name="test_model",
            model_version="test_version",
            features=["user_id", "available_boostpark_location"],
            item_type="test_type",
            num_pick=1,
        )

        results = mbm.predict(
            [
                "1234567890",
                {
                    "available_benefit": ["2234", "2308", "2306", "2286"],
                    "boostpark": "테스트",
                    "context": "로케이션",
                    "location_score": 0.5714,
                },
            ]
        )
        assert len(results["items"]) == 1

        item = results["items"][0]
        assert item["id"] in ["2234", "2308", "2306", "2286"]
        assert item["name"] == "#"
        assert item["type"] == "test_type"
        assert item["props"]["boostpark"] == "테스트"
        assert item["props"]["location_context"] == "로케이션"

    def test_001_mbr_baseline_2(self):
        mbm = MbrBoostparkBaselineModel(
            model_name="test_model",
            model_version="test_version",
            features=["user_id", "available_boostpark_location"],
            item_type="test_type",
            num_pick=2,
        )

        results = mbm.predict(
            [
                "1234567890",
                {
                    "available_benefit": ["2234", "2308", "2306", "2286"],
                    "boostpark": "테스트",
                    "context": "로케이션",
                    "location_score": 0.5714,
                },
            ]
        )
        assert len(results["items"]) == 2

        item1 = results["items"][0]
        assert item1["id"] in ["2234", "2308", "2306", "2286"]
        assert item1["name"] == "#"
        assert item1["type"] == "test_type"
        assert item1["props"]["boostpark"] == "테스트"
        assert item1["props"]["location_context"] == "로케이션"

        item2 = results["items"][1]
        assert item2["id"] in ["2234", "2308", "2306", "2286"]
        assert item2["name"] == "#"
        assert item2["type"] == "test_type"
        assert item2["props"]["boostpark"] == "테스트"
        assert item2["props"]["location_context"] == "로케이션"

        assert item1["props"]["score"] >= item2["props"]["score"]

    def test_002_mbr_baseline_invalid(self):
        mbm = MbrBoostparkBaselineModel(
            model_name="test_model",
            model_version="test_version",
            features=["user_id", "available_boostpark_location"],
            item_type="test_type",
            num_pick=1,
        )

        with pytest.raises(MLSModelError):
            mbm.predict(
                [
                    "1234567890",
                    {
                        "boostpark": "테스트",
                        "context": "로케이션",
                        "location_score": 0.5714,
                    },
                ]
            )

        with pytest.raises(MLSModelError):
            mbm.predict(
                [
                    "1234567890",
                    {
                        "available_benefit": [],
                        "boostpark": "테스트",
                        "context": "로케이션",
                        "location_score": 0.5714,
                    },
                ]
            )

        with pytest.raises(MLSModelError):
            mbm.predict(
                [
                    "1234567890",
                    {
                        "available_benefit": ["2234", "2308", "2306", "2286"],
                        "context": "로케이션",
                        "location_score": 0.5714,
                    },
                ]
            )

        with pytest.raises(MLSModelError):
            mbm.predict(
                [
                    "1234567890",
                    {
                        "available_benefit": ["2234", "2308", "2306", "2286"],
                        "boostpark": "테스트",
                        "location_score": 0.5714,
                    },
                ]
            )



import pytest

from sktmls.models import MLSModelError
from sktmls.models.contrib import MbrInterestAllBaselineModel


class TestMbrInterestAllBaselineModel:
    def test_000_mbr_valid_baseline_1(self):
        mbm = MbrInterestAllBaselineModel(
            model_name="test_model",
            model_version="test_version",
            features=[
                "user_id",
                "prefer_category",
                "available_benefit_activity",
                "available_benefit_asianfood",
                "available_benefit_beauty",
            ],
            categories=["mbr_activity", "mbr_asianfood", "mbr_beauty"],
            item_type="test_type",
            context_id="context",
            num_pick=1,
        )

        results = mbm.predict(["1234567890", ["mbr_beauty"], ["a1", "a2", "a3", "a4"], None, ["b1", "b2", "3", "4"]])
        assert len(results["items"]) == 1

        item = results["items"][0]
        assert item["id"] in ["b1", "b2", "3", "4"]
        assert item["type"] == "test_type"
        assert item["props"]["context_id"] == "context"
        assert item["props"]["prefer_category_yn"] == "Y"
        assert float(item["props"]["score"]) > 1.0

    def test_001_mbr_valid_baseline_2(self):
        mbm = MbrInterestAllBaselineModel(
            model_name="test_model",
            model_version="test_version",
            features=[
                "user_id",
                "prefer_category",
                "available_benefit_activity",
                "available_benefit_asianfood",
                "available_benefit_beauty",
            ],
            categories=["mbr_activity", "mbr_asianfood", "mbr_beauty"],
            item_type="test_type",
            context_id="context",
            num_pick=2,
        )

        results = mbm.predict(["1234567890", ["mbr_beauty"], ["a1", "a2", "a3", "a4"], None, ["b1", "b2", "3", "4"]])
        assert len(results["items"]) == 2

        item1 = results["items"][0]
        assert item1["id"] in ["b1", "b2", "3", "4"]
        assert item1["type"] == "test_type"
        assert item1["props"]["context_id"] == "context"
        assert item1["props"]["prefer_category_yn"] == "Y"
        assert float(item1["props"]["score"]) > 1.0

        item2 = results["items"][1]
        assert item2["id"] in ["b1", "b2", "3", "4"]
        assert item2["type"] == "test_type"
        assert item2["props"]["context_id"] == "context"
        assert item2["props"]["prefer_category_yn"] == "Y"
        assert float(item2["props"]["score"]) > 1.0

        assert item1["props"]["score"] > item2["props"]["score"]

    def test_002_mbr_valid_baseline_2_prefer_none(self):
        mbm = MbrInterestAllBaselineModel(
            model_name="test_model",
            model_version="test_version",
            features=[
                "user_id",
                "prefer_category",
                "available_benefit_activity",
                "available_benefit_asianfood",
                "available_benefit_beauty",
            ],
            categories=["mbr_activity", "mbr_asianfood", "mbr_beauty"],
            context_id="context",
            item_type="test_type",
            num_pick=2,
        )

        results = mbm.predict(["1234567890", ["mbr_beauty"], ["a1", "a2", "a3", "a4"], None, None])
        assert len(results["items"]) == 2

        item1 = results["items"][0]
        assert item1["id"] in ["a1", "a2", "a3", "a4"]
        assert item1["type"] == "test_type"
        assert item1["props"]["context_id"] == "context"
        assert item1["props"]["prefer_category_yn"] == "N"
        assert float(item1["props"]["score"]) < 1.0

        item2 = results["items"][1]
        assert item2["id"] in ["a1", "a2", "a3", "a4"]
        assert item2["type"] == "test_type"
        assert item2["props"]["context_id"] == "context"
        assert item2["props"]["prefer_category_yn"] == "N"
        assert float(item2["props"]["score"]) < 1.0

        assert item1["props"]["score"] > item2["props"]["score"]

    def test_003_mbr_valid_baseline_2_all_none(self):
        mbm = MbrInterestAllBaselineModel(
            model_name="test_model",
            model_version="test_version",
            features=[
                "user_id",
                "prefer_category",
                "available_benefit_activity",
                "available_benefit_asianfood",
                "available_benefit_beauty",
            ],
            categories=["mbr_activity", "mbr_asianfood", "mbr_beauty"],
            context_id="context",
            item_type="test_type",
            num_pick=1,
        )

        results = mbm.predict(["1234567890", ["mbr_beauty"], None, None, None])
        assert len(results["items"]) == 0

    def test_004_mbr_valid_baseline_invalid(self):
        mbm = MbrInterestAllBaselineModel(
            model_name="test_model",
            model_version="test_version",
            features=[
                "user_id",
                "prefer_category",
                "available_benefit_activity",
                "available_benefit_asianfood",
                "available_benefit_beauty",
            ],
            categories=["mbr_activity", "mbr_asianfood", "mbr_beauty"],
            item_type="test_type",
            context_id="context",
            num_pick=1,
        )

        with pytest.raises(MLSModelError):
            mbm.predict(["1234567890", ["mbr_beauty"], ["a1", "a2", "a3", "a4"], None, "good"])

        with pytest.raises(MLSModelError):
            mbm.predict(["1234567890", "good", ["a1", "a2", "a3", "a4"], None, ["b1", "b2", "3", "4"]])
